# -*- coding: utf-8 -*-
#
# Copyright (c) 2021, HILSTER GmbH, Germany - https://hilster.io
# All rights reserved.
#

import serial
from htf.powersupply.core import TimeoutException


class MansonHCS3xxx:
    """
    ``MansonHCS3xxx`` is a hardware abstraction to the Manson HCS-3xxx power supply family using a UART to control it.

    ``MansonHCS3xxx`` uses ``DelegatorMixin`` to automatically set and read it's members `voltage` and `current`.
    """

    def __init__(self, port,
                 baudrate=9600,
                 parity=serial.PARITY_NONE,
                 stopbits=serial.STOPBITS_ONE,
                 timeout=0.5,
                 decimal_places_for_current=1,
                 debug=False):
        """
        Args:
            port (str): the comport to be used, eg. "COM3", 3 or "/dev/ttyUSB0", etc.
            decimal_places_for_current=1 (int): the number of decimal places for the current.
                HCS-3100, 3150, 3200 and 3202 have one decimal place. HCS-3102, 3014 and 3204 have two decimal places.
            debug=False (bool): if set to `True` debugging is enabled
        """

        # open the serial port
        try:
            self.serial = serial.serial_for_url(
                port,
                baudrate=baudrate,
                parity=parity,
                stopbits=stopbits,
                timeout=timeout)  # 9600, 8, N, 1

            self.debug = debug
            self.decimal_places_for_current = decimal_places_for_current
        except Exception as e:
            if hasattr(self, "serial"):
                self.serial.close()
            raise

    # properties
    @property
    def voltage(self):
        return self.get_voltage()

    @voltage.setter
    def voltage(self, value):
        self.set_voltage(value)

    @property
    def current(self):
        return self.get_current()

    @current.setter
    def current(self, value):
        self.set_current(value)

    def query(self, query, answer_lines=1):
        """
        Send a command and wait for the answer.

        Args:
            query (str): the query to be sent.
            answer_lines=1 (int): the number of lines to be received.

        Returns:
            str: the answer line(s).

        Raises:
            manson_hcs_3xxx.TimeoutException: if a communication time-out occurred.
        """
        if not isinstance(query, bytes):  # Python 3 hack
            query = query.encode()

        request = query + b"\x0D"
        if self.debug:
            print("sending:")
            print(request)
        self.serial.write(request)

        response = []

        def readline():
            ch = line = self.serial.read(1)
            while ch not in [b"\x0D"]:
                ch = self.serial.read(1)
                line += ch
                if self.debug and ch not in [None, b""]:
                    print("bytes in receive buffer:", self.serial.inWaiting())
                    print(hex(ord(ch)))

            return line[:-1]

        for i in range(answer_lines):
            answer = readline()
            if self.debug:
                print("line:", answer)
            if answer not in [None, ""]:
                response.append(answer)
            else:
                raise TimeoutException("Device does not respond!")

        if response[-1] != b"OK":
            return None

        return response[:-1]

    def get_preset_maximum_voltage_and_current(self):
        """
        Get the preset maximum voltage and current.

        Returns:
            dict: a dictionary containing "voltage" and "current".
        """

        response = self.query(query=b"GMAX", answer_lines=2)

        voltage = int(response[0][:3]) / 10.0
        current = int(response[0][3:3 + 3]) / 10.0

        return {"voltage": voltage, "current": current}

    getPresetMaximumVoltageAndCurrent = get_preset_maximum_voltage_and_current

    def get_preset_voltage_and_current(self):
        """
        Get the preset voltage and current.

        Returns:
            dict: a dictionary containing "voltage" and "current".
        """

        response = self.query(query=b"GETS", answer_lines=2)

        voltage = int(response[0][:3]) / 10.0
        current = int(response[0][3:3 + 3]) / 10.0

        return {"voltage": voltage, "current": current}

    getPresetVoltageAndCurrent = get_preset_voltage_and_current

    def get_display_voltage_current_and_status(self):
        """
        Get the display voltage, current and status.

        Returns:
            dict: a dictionary containing "voltage", "current" and "status".
        """
        response = self.query(query=b"GETD", answer_lines=2)

        voltage = int(response[0][:4]) / 100.0
        current = int(response[0][4:]) / (10.0 ** self.decimal_places_for_current)

        tmp = response[0]
        if not isinstance(tmp, str):
            tmp = tmp.decode()

        status = ["CV", "CC"][int(tmp[-1])]

        return {"voltage": voltage, "current": current, "status": status}

    getDisplayVoltageCurrentAndStatus = get_display_voltage_current_and_status

    def get_preset_maximum_voltage(self):
        """
        Get preset maximum voltage for current preset.

        Returns:
            int: the preset maximum voltage.
        """
        response = self.query(query=b"GOVP", answer_lines=2)

        voltage = int(response[0]) / 10.0

        return voltage

    getPresetMaximumVoltage = get_preset_maximum_voltage

    def get_preset_maximum_current(self):
        """
        Get preset maximum current for current preset.

        Returns:
            int: the preset maximum current.
        """
        response = self.query(query=b"GOCP", answer_lines=2)

        current = int(response[0][:3]) / 10.0

        return current

    getPresetMaximumCurrent = get_preset_maximum_current

    def preset_maximum_voltage(self, voltage):
        """
        Set maximum voltage for current preset.
        """
        command = b"SOVP" + self._get_voltage_string(voltage)
        self.query(command)

    presetMaximumVoltage = preset_maximum_voltage

    def preset_maximum_current(self, current):
        """
        Set maximum voltage for current preset.
        """
        command = b"SOCP" + self._get_current_string(current)
        self.query(command)

    presetMaximumCurrent = preset_maximum_current

    def store_presets(self, presets):
        """
        Store three voltage-current presets.

        Args:
            presets (dict): a dictionary containing the key P1, P2 and P3. Each key contains "voltage" and
             "current" keys.
        """

        command = b"PROM"
        for preset in range(3):
            command += self._get_voltage_string(presets["P%1i" % (preset + 1)]["voltage"])
            command += self._get_current_string(presets["P%1i" % (preset + 1)]["current"])

        self.query(command)

    storePresets = store_presets

    def get_presets(self):
        """
        Get three voltage-current presets.

        Returns:
            dict: a dictionary containing "voltage", "current" and "P<preset>".
        """
        response = self.query(query=b"GETM", answer_lines=4)

        preset_number = 1

        presets = {}

        for line in response:
            voltage = int(line[:3]) / 10.0
            current = int(line[3:]) / (10.0 ** self.decimal_places_for_current)

            preset = {"voltage": voltage, "current": current}
            presets["P%i" % preset_number] = preset
            preset_number += 1

        return presets

    getPresets = get_presets

    def load_preset(self, preset):
        """
        Load preset ``preset``.

        Args:
            preset (int): may be 0, 1, or 2.
        """
        self.query(b"RUNM%1i" % int(preset))

    loadPreset = load_preset

    def enable_output(self):
        """
        Enable output.
        """
        self.query(b"SOUT0")

    enableOutput = enable_output

    def disable_output(self):
        """
        Disable output.
        """
        self.query(b"SOUT1")

    disableOutput = disable_output

    @staticmethod
    def _get_voltage_string(voltage):
        return b"%03i" % int(float(voltage) * 10.0)

    def set_voltage(self, voltage):
        """
        Set the output voltage to ``voltage``.

        Args:
            voltage (int, float or str): the voltage in units of volts as int(), float() or string, eg. 23, 23.0,
                "23.0" or even "23.0 V" work.
                `voltage` may have on decimal point. If `voltage` is ``None`` the output is disabled.
        """
        if voltage is None:
            self.disableOutput()
        else:
            command = b"VOLT" + self._get_voltage_string(voltage)
            self.query(command)
            self.enable_output()

    setVoltage = set_voltage

    def get_voltage(self):
        """
        Read current voltage from device and return the result.

        Returns:
            float: the current voltage.
        """
        data = self.get_display_voltage_current_and_status()
        return data["voltage"]

    getVoltage = get_voltage

    def _get_current_string(self, current):
        i = float(current)
        i *= 10.0 ** self.decimal_places_for_current

        format_ = b"%%0%ii" % (2 + self.decimal_places_for_current)
        return format_ % int(i)

    def set_current(self, current):
        """
        Set the output current to ``current``.

        Args:
            current (int, float or str): the current in units of amperes as int(), float() or string, eg. 0, 0.5,
             "0.5" or even "0.5 mA" work.
        """

        command = b"CURR" + self._get_current_string(current)
        self.query(command)

    setCurrent = set_current

    def get_current(self):
        """
        Read the current current from device and return the result.

        Returns:
            float: the current current.
        """
        data = self.get_display_voltage_current_and_status()
        return data["current"]

    getCurrent = get_current

    def close(self):
        """
        Close the serial member. This method is automatically called when :func:`__del__` is called.
        """
        if hasattr(self, "serial"):
            self.serial.close()

    def __del__(self):
        """
        Destructor of an ``MansonHCS3xxx`` instance.
        This method calls :func:`close` to automatically close the serial instance.
        """
        self.close()
